#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Rotation Gate Operation
"""
import importlib
from typing import List, Optional, TYPE_CHECKING

import numpy

from Quanlse.QOperation import QOperation
from Quanlse.QRegPool import QRegStorage

if TYPE_CHECKING:
    from Quanlse.QOperation import RotationArgument, OperationFunc


class RotationGateOP(QOperation):
    """
    Rotation gate

    Use rotation parameters to create the quantum operators
    """
    argumentList = None  # type: List['RotationArgument']
    uGateArgumentList = None  # type: List['RotationArgument']

    def __init__(self, gate: str, bits: int,
                 angleList: List['RotationArgument'],
                 uGateArgumentList: List['RotationArgument']) -> None:
        """
        Constructor for RotaionGateOP class
        """
        super().__init__(gate, bits)
        self.argumentList = angleList
        self.uGateArgumentList = uGateArgumentList

    def __call__(self, *qRegList: QRegStorage, gateTime: Optional[float] = None) -> None:
        self._op(list(qRegList))

    def generateMatrix(self) -> numpy.ndarray:
        """
        Generate the matrix of rotation gate.
        """
        self._generateUMatrix()
        return self._matrix

    def _u3Matrix(self, theta: float, phi: float, lamda: float) -> numpy.ndarray:
        """
        Generate a single-qubit rotation gate with 3 angles

        :param theta: angle
        :param phi: angle
        :param lamda: angle
        :return: U3 matrix
        """

        self._matrix = numpy.array([[numpy.cos(theta / 2.0), -numpy.exp(1j * lamda) * numpy.sin(theta / 2.0)],
                                    [numpy.exp(1j * phi) * numpy.sin(theta / 2.0),
                                     numpy.exp(1j * lamda + 1j * phi) * numpy.cos(theta / 2.0)]])
        return self._matrix

    def _u2Matrix(self, phi: float, lamda: float) -> numpy.ndarray:
        """
        Generate a single-qubit rotation gate with 2 angles

        :param phi: angle
        :param lamda: angle
        :return: U2 matrix
        """

        self._matrix = (1 / numpy.sqrt(2)) * numpy.array([[1, -numpy.exp(1j * lamda)],
                                                          [numpy.exp(1j * phi), numpy.exp(1j * (phi + lamda))]])
        return self._matrix

    def _u1Matrix(self, lamda: float) -> numpy.ndarray:
        """
        Generate a single-qubit rotation gate along the Z-axis

        :param lamda: angle
        :return: U1 matrix
        """

        self._matrix = numpy.array([[1, 0],
                                    [0, numpy.exp(1j * lamda)]])
        return self._matrix

    def _cu3Matrix(self, theta: float, phi: float, lamda: float) -> numpy.ndarray:
        self._matrix = numpy.kron(numpy.eye(2),
                                  numpy.array([[1, 0],
                                               [0, 0]])
                                  ) + \
                       numpy.kron(self._u3Matrix(theta, phi, lamda),
                                  numpy.array([[0, 0],
                                               [0, 1]])
                                  )
        return self._matrix

    def _generateUMatrix(self) -> None:
        """
        Generate U matrix

        :return: None
        """
        uGateArgumentCount = len(
            [value for value in self.uGateArgumentList if isinstance(value, (float, int))])
        if uGateArgumentCount != len(self.uGateArgumentList):
            pass  # has parameter
        elif uGateArgumentCount == 3:
            self._u3Matrix(*self.uGateArgumentList)
        elif uGateArgumentCount == 2:
            self._u2Matrix(*self.uGateArgumentList)
        elif uGateArgumentCount == 1:
            self._u1Matrix(*self.uGateArgumentList)

    def _generateCUMatrix(self) -> None:
        """
        Return CU matrix

        :return: None
        """
        uGateArgumentCount = len([value for value in self.uGateArgumentList if isinstance(value, (float, int))])
        if uGateArgumentCount != len(self.uGateArgumentList):
            pass  # has parameter
        elif uGateArgumentCount == 3:
            self._cu3Matrix(*self.uGateArgumentList)
        # elif uGateArgumentCount == 2:
        #     self._cu2Matrix(*self.uGateArgumentList)
        # elif uGateArgumentCount == 1:
        #     self._cu1Matrix(*self.uGateArgumentList)


def U(theta: 'RotationArgument',
      phi: Optional['RotationArgument'] = None,
      lamda: Optional['RotationArgument'] = None) -> 'OperationFunc':
    """
    U Gate

    Generate a single-qubit U1 (or U2 or U3) gate according to the number of angles.
    """
    uGateArgumentList = angleList = [value for value in [theta, phi, lamda] if value is not None]
    gate = RotationGateOP('U', 1, angleList, uGateArgumentList)
    gate.generateMatrix = gate._generateUMatrix()
    return gate


def RX(theta: 'RotationArgument') -> 'OperationFunc':
    """
    RX Gate

    Single-qubit rotation about the X-axis.

    According to the relation: U3(theta, -pi/2, pi/2) = RX(theta)
    """
    angleList = [theta]
    uGateArgumentList = [theta, -numpy.math.pi / 2, numpy.math.pi / 2]
    gate = RotationGateOP('RX', 1, angleList, uGateArgumentList)
    gate.generateMatrix = gate._generateUMatrix()
    return gate


def RY(theta: 'RotationArgument') -> 'OperationFunc':
    """
    RY Gate

    Single-qubit rotation about the Y-axis.

    According to the relation: U3(theta, 0, 0) = RY(theta)
    """
    angleList = [theta]
    uGateArgumentList = [theta, 0, 0]
    gate = RotationGateOP('RY', 1, angleList, uGateArgumentList)
    gate.generateMatrix = gate._generateUMatrix()
    return gate


def RZ(lamda: 'RotationArgument') -> 'OperationFunc':
    """
    RZ Gate

    Single-qubit rotation about the Z-axis.

    According to the relation: U3(0, 0, lamda) = RZ(lamda)
    """
    angleList = [lamda]
    uGateArgumentList = [0, 0, lamda]
    gate = RotationGateOP('RZ', 1, angleList, uGateArgumentList)
    gate.generateMatrix = gate._generateUMatrix()
    return gate


def VZ(lamda: 'RotationArgument') -> 'OperationFunc':
    """
    VZ Gate

    Single-qubit rotation about the Z-axis by phase shift.

    According to the relation: U3(0, 0, lamda) = VZ(lamda)
    """
    angleList = [lamda]
    uGateArgumentList = [0, 0, lamda]
    gate = RotationGateOP('VZ', 1, angleList, uGateArgumentList)
    gate.generateMatrix = gate._generateUMatrix()
    return gate



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bloch Sphere Visualization Module
"""

import time
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import animation
import numpy as np
from typing import Union

from Quanlse.QOperation import Error


def plotBloch(coordinates: list = None, xyzData: list = None, color: str ='Purple', mode: str ='animate',
              interval: float = None, save: bool = False, size: Union[int, float] = 2, title: str = '',
              fName: str = None, fps: float = 20):
    """
    This function plots Bloch sphere - supports 'animate' 'vector' and 'scatter' mode.

    :param coordinates: The list of Cartesian coordinates of n data points to plot.
    :param xyzData: The list of [XS YS ZS] of n data points to plot.
    :param color: The color of the point.
    :param mode: support 'animate' 'vector' and 'scatter' modes.
    :param interval: Delay between frames in ms.
    :param save: Whether to save animation as a gif file.
    :param size: Size of points in scatter mode.
    :param title: The title of the figure.
    :param fName: The name of the file saved.
    :param fps: The fps of gif file to save.
    """

    if coordinates is not None:
        coordinates = np.array(coordinates)
        coordinates[:, [0, 1]] = -1 * coordinates[:, [1, 0]]
        data = np.array(coordinates).T

    elif xyzData is not None:
        data = np.array(xyzData)
        data[[0, 1], :] = -1 * data[[1, 0], :]
        coordinates = data.T

    else:
        raise Error.ArgumentError('xyzData and coordinate unfilled.')

    # create figure
    fig = plt.figure(figsize=[6, 6])
    ax = Axes3D(fig)

    # Set X range
    ax.set_xlim3d([-1, 1])
    ax.set_xlabel('X')

    # Set Y range
    ax.set_ylim3d([-1, 1])
    ax.set_ylabel('Y')

    # Set Z range
    ax.set_zlim3d([-1, 1])
    ax.set_zlabel('Z')

    pNumber = 100

    # Plot Bloch Sphere
    u = np.linspace(0, 2 * np.pi, pNumber)
    v = np.linspace(0, np.pi, pNumber)

    x = 1 * np.outer(np.cos(u), np.sin(v))
    y = 1 * np.outer(np.sin(u), np.sin(v))
    z = 1.3 * np.outer(np.ones(np.size(u)), np.cos(v))

    ax.plot_surface(x, y, z, rstride=4, cstride=4, color='ghostwhite', linewidth=0, alpha=0.2)
    ax.plot(np.sin(u), np.cos(u), 0, color='grey')
    ax.plot([0] * pNumber, np.sin(u), 1.3 * np.cos(u), color='grey')

    # Hide grid lines
    ax.grid(False)

    # Hide default axis
    plt.axis('off')

    # Plot Axis and Axis Label
    ax.quiver(-1.2, 0, 0, 1.2, 0, 0, length=2, normalize=False,
              arrow_length_ratio=0.03, lw=1.5, color='black')
    ax.quiver(0, 1.2, 0, 0, -1.2, 0, length=2, normalize=False,
              arrow_length_ratio=0.03, lw=1.5, color='black')
    ax.quiver(0, 0, -1.2 * 1.3, 0, 0, 1.2 * 1.3, length=2, normalize=False,
              arrow_length_ratio=0.03, lw=2.5, color='black')
    ax.text(0, 0, 1.3 * 1.3, s='Z', weight='heavy')
    ax.text(0, -1.4, 0, s='x', weight='heavy')
    ax.text(1.3, 0, 0, s='y', weight='heavy')

    if mode == 'animate':

        data[2] = data[2] * 1.3

        line = plt.plot(data[0], data[1], data[2], lw=3)[0]

        def updateTrajectory(num, data, line):
            line.set_data(data[0:2, :num])
            line.set_3d_properties(data[2, :num])
            return line

        def update(num, data, line):
            update.vec.remove()
            update.vec = ax.quiver(0, 0, 0, data[0][num], data[1][num], data[2][num],
                                   normalize=False, arrow_length_ratio=0.05, lw=2, color=color)
            line = updateTrajectory(num, data, line)
            return line

        numData = len(coordinates)

        if interval is None:
            interval = 3 / numData

        update.vec = ax.quiver(0, 0, 0, 0, 0, 0)

        line_ani = animation.FuncAnimation(fig=fig, func=update, frames=numData,
                                           fargs=(data, line), interval=interval)

        plt.show()
        if save:
            writerGif = animation.PillowWriter(fps=fps)
            now = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime())
            line_ani.save(now + title + '.gif', writer=writerGif)

    elif mode == 'vector':
        # Vector mode
        coordinates = np.array(data).T
        for dat in coordinates:
            ax.quiver(0, 0, 0, dat[0], dat[1], dat[2] * 1.3,
                      normalize=False, arrow_length_ratio=0.1, lw=2, color=color)
        plt.title(label=title)
        if save:
            if fName is None:
                fName = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime()) + title + '.png'
            plt.savefig(fName)
        plt.show()

    elif mode == 'scatter':
        # Scatter mode
        ax.scatter(data[0], data[1], data[2] * 1.3, marker='o', color=color, s=size)

        plt.title(label=title)
        if save:
            if fName is None:
                fName = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime()) + title + '.png'
            plt.savefig(fName)
        plt.show()


def rho2Coordinate(rho: np.ndarray) -> list:
    """
    Convert a density matrix to the list of Cartesian coordinates.

    :param rho: The density matrix.
    :return: The list of Cartesian coordinates for given density matrix.
    """

    x1 = 2 * rho[0][1].real
    x2 = 2 * rho[1][0].imag
    x3 = (rho[0][0] - rho[1][1]).real

    return [x1, x2, x3]



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pipeline class for Quanlse Scheduler.
"""

import copy
from typing import List, Union, Dict, Any, Callable

from Quanlse.QWaveform import QJob
from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.Scheduler.GatePulsePair import GatePulsePair


class SchedulerProcess:
    """
    Scheduler process for SchedulerPipeline instances to perform scheduling strategy.

    :param title: a user-given title to the pi object
    :param processFunc: Callable function to perform scheduling strategies
    """

    def __init__(self, title: str, processFunc: Callable):
        """
        Constructor for the SchedulerProcess class.
        """
        self.title = title
        self._processFunc = processFunc
        self._conf = {}  # type: Dict[str, Any]

    def __call__(self, job: QJob, scheduler: 'Scheduler') -> QJob:
        """
        Call the scheduling function.
        """
        if self._processFunc is None:
            raise ArgumentError("Scheduling function is not set.")
        return self._processFunc(job=job, scheduler=scheduler)


class SchedulerPipeline:
    """
    SchedulerPipeline instance is a sequence of SchedulerProcess to perform scheduling strategies.

    :param title: a user-given title to the pi object
    """

    def __init__(self, title: str = ""):
        """
        Initializing
        """
        self._title = title
        self._pipeline = []  # type: List[SchedulerProcess]

    def __call__(self, job: QJob, scheduler: 'Scheduler') -> QJob:
        """
        Call the instance and run the pipeline in order,
        and output a QJob instance.
        """
        for idx, process in enumerate(self._pipeline):
            if isinstance(process, Callable):
                job = process(job, scheduler)
            else:
                raise ArgumentError("Scheduler process is not callable.")
        return job

    @property
    def pipeline(self) -> List[SchedulerProcess]:
        """
        Return jobs.
        """
        return self._pipeline

    def addPipelineJob(self, process: Union[SchedulerProcess, List[SchedulerProcess]]) -> None:
        """
        Add pipeline job in to the job list
        """
        if isinstance(process, List):
            for pl in process:
                self._pipeline.append(pl)
        elif isinstance(process, SchedulerProcess):
            self._pipeline.append(process)

    def setPipelineJob(self, process: Union[SchedulerProcess, List[SchedulerProcess]]) -> None:
        """
        Remove all existed pipeline jobs and add the new job in to the job list
        """
        del self._pipeline
        self._pipeline = []
        if isinstance(process, List):
            for pl in process:
                self._pipeline.append(pl)
        elif isinstance(process, SchedulerProcess):
            self._pipeline.append(process)

    def __add__(self, other: Union[SchedulerProcess, List[SchedulerProcess], 'SchedulerPipeline']) \
            -> 'SchedulerPipeline':
        """ Add  """
        if isinstance(other, SchedulerProcess):
            # Add a Pipeline job in to the list
            self.addPipelineJob(other)
            return self
        elif isinstance(other, SchedulerPipeline):
            for job in other.pipeline:
                self.addPipelineJob(copy.deepcopy(job))
            return self


def findMaxT(layer: list, scheduler: 'Scheduler') -> list:
    """
    Find maximum time for each layer

    :param layer: circuit list
    :param scheduler: a Scheduler object
    :return: a list containing the maximum time in each layer
    """
    maxDepth = [0 for _ in range(len(layer))]
    for layIdx, lay in enumerate(layer):
        for gateIdx in range(len(lay)):
            if lay[gateIdx] is not None:
                if scheduler.savePulse:
                    someJob = scheduler.findGatePulsePair(lay[gateIdx])
                else:
                    job = scheduler.pulseGenerator(lay[gateIdx], scheduler)
                    someJob = GatePulsePair(lay[gateIdx], job)
                if someJob.t > maxDepth[layIdx]:
                    maxDepth[layIdx] = someJob.t
    return maxDepth


def toLayer(layer: list, scheduler: 'Scheduler') -> None:
    """
    Turn gates in a Scheduler object into layers

    :param layer: empty list to store circuit
    :param scheduler: a Scheduler object
    :return: None
    """
    for gate in scheduler.circuit:
        if len(gate.qRegIndexList) == 1:
            layer[gate.qRegIndexList[0]].append(gate)
        elif len(gate.qRegIndexList) >= 2:
            fillUp(layer, gate.qRegIndexList)
            for qubit in gate.qRegIndexList:
                layer[qubit].append(gate)


def fillUp(layer: list, qubits: list = None) -> None:
    """
    Fill up till each list is of equal size

    :param layer: circuit list
    :param qubits: qubit indexes to fill
    :return: None
    """
    if qubits is None:
        qubits = range(len(layer))

    lens = [len(layer[qubit]) for qubit in qubits]
    for qubit in qubits:
        if max(lens) - len(layer[qubit]) != 0:
            N = [None] * (max(lens) - len(layer[qubit]))
            layer[qubit] = layer[qubit] + N


def addToJob(layer: list, scheduler: 'Scheduler', maxT: list, job: QJob) -> None:
    """
    Add the circuit in the given layer to scheduler.job

    :param layer: circuit list
    :param scheduler: original Scheduler object
    :param maxT: a list generated by findMaxt()
    :param job: the QJob object to add the pulses to
    """

    t0 = 0
    padding = scheduler.padding

    # for each layer
    for i in range(len(layer)):
        temp = []

        # for each qubit
        for j in layer[i]:

            # dont add the same gate multiple times
            flag = False
            for tp in temp:
                if tp == j and j is not None:
                    flag = True
            if j is not None and flag is False:
                if scheduler.savePulse:
                    obj = scheduler.findGatePulsePair(j)
                else:
                    jobObj = scheduler.pulseGenerator(j, scheduler)
                    obj = GatePulsePair(j, jobObj)
                if t0 == 0:
                    job.appendJob(obj.pulses, t0)
                else:
                    if (t0 - padding * 2) < 0:
                        raise ValueError("Error: buffer time exceeds pulse length.")
                    job.appendJob(obj.pulses, t0 - padding * (i + 1))

            temp.append(j)

        if t0 == 0:
            t0 += maxT[i] - padding
        else:
            t0 += maxT[i] - padding * 2


def reverseToLayer(layer: list, scheduler: 'Scheduler'):
    """
    Turn gates in Scheduler into layers
    """
    for gate in reversed(scheduler.circuit):
        if len(gate.qRegIndexList) == 1:
            layer[gate.qRegIndexList[0]].insert(0, gate)
        elif len(gate.qRegIndexList) >= 2:
            reverseFillUp(layer, gate.qRegIndexList)
            for qubit in gate.qRegIndexList:
                layer[qubit].insert(0, gate)


def reverseFillUp(layer: list, qubits=None):
    """
    Fill up till the list is of equal size
    """

    if qubits is None:
        qubits = range(len(layer))

    lens = [len(layer[qubit]) for qubit in qubits]
    for qubit in qubits:
        if max(lens) - len(layer[qubit]) != 0:
            N = [None] * (max(lens) - len(layer[qubit]))
            layer[qubit] = N + layer[qubit]


def leftAlignSingleGates(layer: list):
    """
    Left align all the single-qubit gates
    """

    for qubit in layer:
        index = len(qubit) - 1
        while index >= 0:
            if qubit[index] is not None and len(qubit[index].qRegIndexList) == 1 \
                    and index - 1 >= 0 and qubit[index - 1] is None:
                qubit[index - 1], qubit[index] = qubit[index], qubit[index - 1]
                if index + 2 < len(qubit):
                    index += 2
            index -= 1
    return layer


def delayFirstGate(layer, gateNumberBeforeMulti):
    """
    Delay the first gates
    """

    for i in range(len(layer[0])):
        tempList = []
        for j in range(gateNumberBeforeMulti[i]):
            if layer[j][i] is not None:
                tempList.append(layer[j][i])
        if len(tempList) != gateNumberBeforeMulti[i]:
            count = len(tempList) - 1
            for num in reversed(range(gateNumberBeforeMulti[i])):
                if count >= 0:
                    layer[num][i] = tempList[count]
                else:
                    layer[num][i] = None
                count -= 1



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Benchmark
"""

from typing import Dict, Any, List
from numpy import ndarray, dot, array

from Quanlse.Utils.Functions import expect, basis
from Quanlse.QHamiltonian import QHamiltonian
from Quanlse.QPlatform.Error import ArgumentError


def evolution(ham: QHamiltonian, stateInitial: ndarray = None, matrix: ndarray = None) -> Dict[str, Any]:
    """
    Return the expectation value of the given matrix for given initial states. The input can be a list
    containing all the matrices and states that the users want to calculate.

    :param ham: Hamiltonian Dictionary
    :param stateInitial: the list of numpy.ndarray that represents an initial state in the Hilbert space
    :param matrix: the list of matrix whose expectation values to be evaluated
    :return: the dictionary that contains all the expectation values for all initial states
    """

    # If the input data is not list, we transform it into a list
    if type(matrix) is ndarray:
        matrixList = [matrix]
    else:
        matrixList = matrix
    if type(stateInitial) is ndarray:
        stateInitial = [stateInitial]

    result = ham.simulate(recordEvolution=True)
    unitaryList = result[0]['evolution_history']
    # x = numpy.linspace(0, ham['circuit']['max_time_ns'], ham['circuit']['max_time_dt'])

    stateList = {}
    # We first calculate all the intermediate states in the process
    for stateIndex, stateItem in enumerate(stateInitial):
        stateListTemp = evolutionList(stateItem, unitaryList)
        stateList[str(stateIndex)] = {}
        stateList[str(stateIndex)]['state_form'] = stateItem
        stateList[str(stateIndex)]['state_evolution_history'] = stateListTemp
        matrixTempDict = {}
        for matrixIndex, matrixItem in enumerate(matrixList):
            matrixTempDict['matrix-' + str(matrixIndex) + '-form'] = matrixItem
            valueList = []
            for stateItemTemp in stateListTemp:
                valueList.append(expect(matrixItem, stateItemTemp))
            matrixTempDict['matrix-' + str(matrixIndex) + '-value'] = valueList
        stateList[str(stateIndex)]['result'] = matrixTempDict
    return stateList


def evolutionList(state: ndarray, unitaryList: List) -> list:
    """
    Return the intermediate states with given initial states and the list of unitary matrices.

    :param state: the numpy.ndarray representing the initial state
    :param unitaryList: the list containing different unitary matrices at different times
    :return stateList: the list of intermediate states for the given unitary matrices
    """
    stateList = []
    for unitaryItem in unitaryList:
        stateList.append(dot(unitaryItem, state))
    return stateList


def stateTruthTable(unitary, qubitNum, sysLevel, initialBasisList=None, finalBasisList=None) -> ndarray:
    """
    Generate the truth table of a quantum gate contains the probability of the system being in each
    possible basis states at the end of an operation for each possible initial state.

    :param unitary: the unitary matrix
    :param qubitNum: the number of qubits
    :param sysLevel: the energy level of the system
    :param initialBasisList: the list initial basis
    :param finalBasisList: the list final basis
    :return: the population matrix
    """
    if not isinstance(sysLevel, int):
        raise ArgumentError('This function currently only supports an integer system level as input.')

    resultMatrix = []
    if initialBasisList is None:
        initialBasisList = range(sysLevel ** qubitNum)
    if finalBasisList is None:
        finalBasisList = range(sysLevel ** qubitNum)
    for state in initialBasisList:
        stateVec = basis(sysLevel ** qubitNum, state)
        stateFinalTemp = unitary @ stateVec
        probVec = []
        for index, item in enumerate(stateFinalTemp):
            if index in finalBasisList:
                probVec.append(abs(item) ** 2)
        resultMatrix.append(probVec)

    return array(resultMatrix)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file aims to collect functions related to the Clifford group.
"""

from typing import List
import numpy as np

from Quanlse.QOperation.FixedGate import FixedGateOP
from Quanlse.QOperation.RotationGate import RX, RY, RZ
from Quanlse.QPlatform.Error import ArgumentError

# The basic elements of Clifford gates
x90p = RX(np.pi / 2).getMatrix()  # Rx+
x90m = RX(-np.pi / 2).getMatrix()  # Rx-
xp = RX(np.pi).getMatrix()  # Rx
y90p = RY(np.pi / 2).getMatrix()  # Ry+
y90m = RY(-np.pi / 2).getMatrix()  # Ry-
yp = RY(np.pi).getMatrix()  # Ry
z90p = RZ(np.pi / 2).getMatrix()  # Rz+
z90m = RZ(-np.pi / 2).getMatrix()  # Rz-
zp = RZ(np.pi).getMatrix()  # Rz
Id = np.array([[1, 0], [0, 1]], dtype=np.complex128)  # Identity

# Generate single-qubit Clifford matrices
C1 = FixedGateOP('C1', 1, Id)
C2 = FixedGateOP('C2', 1, xp)
C3 = FixedGateOP('C3', 1, yp)
C4 = FixedGateOP('C4', 1, xp @ yp)
C5 = FixedGateOP('C5', 1, x90p @ y90p)
C6 = FixedGateOP('C6', 1, x90p @ y90m)
C7 = FixedGateOP('C7', 1, x90m @ y90p)
C8 = FixedGateOP('C8', 1, x90m @ y90m)
C9 = FixedGateOP('C9', 1, y90p @ x90p)
C10 = FixedGateOP('C10', 1, y90p @ x90m)
C11 = FixedGateOP('C11', 1, y90m @ x90p)
C12 = FixedGateOP('C12', 1, y90m @ x90m)
C13 = FixedGateOP('C13', 1, x90p)
C14 = FixedGateOP('C14', 1, x90m)
C15 = FixedGateOP('C15', 1, y90p)
C16 = FixedGateOP('C16', 1, y90m)
C17 = FixedGateOP('C17', 1, x90m @ y90p @ x90p)
C18 = FixedGateOP('C18', 1, x90m @ y90m @ x90p)
C19 = FixedGateOP('C19', 1, xp @ y90p)
C20 = FixedGateOP('C20', 1, xp @ y90m)
C21 = FixedGateOP('C21', 1, yp @ x90p)
C22 = FixedGateOP('C22', 1, yp @ x90m)
C23 = FixedGateOP('C23', 1, x90p @ y90p @ x90p)
C24 = FixedGateOP('C24', 1, x90m @ y90p @ x90m)

# Set the single-qubit Clifford gate set
clifford1q = [C1, C2, C3, C4, C5, C6, C7, C8, C9, C10,
              C11, C12, C13, C14, C15, C16, C17, C18, C19, C20,
              C21, C22, C23, C24]


def randomClifford(qubits: int = 1, size: int = 1, seed: int = None) -> List[FixedGateOP]:
    r"""
    Randomly generate `size` number amount of Clifford operators,
    whose number of qubit(s) is given by `qubits`.

    :param qubits: the number of qubits of each Clifford operator
    :param size: the number of Clifford operators generated
    :param seed: the seed used for randomly generating Clifford operators
    :return: A list of generated Clifford operators in U3 representation.
            Note that the return type is always a list.
    """
    if qubits != 1:
        raise ArgumentError(f'in random_Clifford(): only support single-qubit Clifford now!')

    if seed is None:
        idx = np.random.randint(0, 24, size)
    else:
        np.random.seed(seed)
        idx = np.random.randint(0, 24, size)

    cliffordRandomList = [clifford1q[idx[i]] for i in range(size)]

    return cliffordRandomList



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Define Error
"""


class Error(Exception):
    """
    The basic class of the user-defined exceptions
    """

    code = 0

    def __init__(self, msg=None):
        self.message = msg


class ArgumentError(Error):
    """
    Arguments related error
    """

    code = 100


class NetworkError(Error):
    """
    Network related error
    """

    code = 200


class RuntimeError(Error):
    """
    Runtime related error
    """

    code = 300


class LogicError(Error):
    """
    Logical error
    """

    code = 400


class TokenError(LogicError):
    """
    Token related error
    """

    code = 401



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Default SchedulerPipeline for QuanlseSchedulerSuperconduct.
"""

from numpy import array

from Quanlse.Scheduler.SchedulerPipeline import SchedulerProcess, toLayer, fillUp, findMaxT, addToJob
from Quanlse.Scheduler import Scheduler
from Quanlse.QWaveform import QJob


def _leftAlignedCore(job: QJob, scheduler: Scheduler) -> QJob:
    """
    The left alignment strategy for scheduling.

    :param job: job to be returned
    :param scheduler: Scheduler object containing the circuit information
    :return: the returned QJob object
    """

    # Initialize layers
    layer = []
    for i in range(scheduler.subSysNum):
        layer.append([])

    # First convert gates in Scheduler to layers
    toLayer(layer, scheduler)
    fillUp(layer)

    # Transpose layers
    layer = array(layer).T.tolist()

    # Find max time for each layer
    maxT = findMaxT(layer, scheduler)

    # clear Waves
    job.clearWaves()
    job.clearCache()

    # add waves to job
    addToJob(layer=layer, scheduler=scheduler, maxT=maxT, job=job)

    job.buildWaveCache()

    return job


leftAligned = SchedulerProcess("LeftAligned", _leftAlignedCore)
"""
A SchedulerProcess instance containing the left-aligned scheduling strategy.
"""



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Default SchedulerPipeline for QuanlseSchedulerSuperconduct.
"""

from numpy import array

from Quanlse.Scheduler.SchedulerPipeline import SchedulerProcess, reverseToLayer, reverseFillUp, \
    leftAlignSingleGates, delayFirstGate, findMaxT, addToJob
from Quanlse.Scheduler import Scheduler
from Quanlse.QWaveform import QJob


def _gatesBeforeMulti(layer, scheduler):
    """
    Find gate number before the first multi-qubit gate
    """
    flag = [True] * scheduler.subSysNum
    gateNumber = [len(layer)] * scheduler.subSysNum
    for i in range(len(layer[0])):
        for j in range(len(layer)):
            if layer[j][i] is not None:
                if len(layer[j][i].qRegIndexList) >= 2 and flag[i] is True:
                    gateNumber[i] = j
                    flag[i] = False
    return gateNumber


def _centerAlignedCore(job: QJob, scheduler: Scheduler) -> QJob:
    """
    The left alignment strategy for scheduling.

    :param job: job to be returned
    :param scheduler: Scheduler object containing the circuit information
    :return: the returned QJob object
    """

    # Initialize layers
    layer = []
    for i in range(scheduler.subSysNum):
        layer.append([])

    # First convert gates in Scheduler to layers
    reverseToLayer(layer, scheduler)
    reverseFillUp(layer)

    # left aligned the single-qubit gates
    leftAlignSingleGates(layer)

    # Transpose layers
    layer = array(layer).T.tolist()

    # find gate number before the first multi-qubit gates
    gatesBeforeMultiQubitGate = _gatesBeforeMulti(layer, scheduler)

    # start the first gate as late as possible
    delayFirstGate(layer, gatesBeforeMultiQubitGate)

    # clear wave and cache
    job.clearWaves()
    job.clearCache()

    # find max time for each layer
    maxi = findMaxT(layer, scheduler)

    # add waves to job
    addToJob(layer=layer, scheduler=scheduler, maxT=maxi, job=job)

    job.buildWaveCache()

    return job


centerAligned = SchedulerProcess("CenterAligned", _centerAlignedCore)
"""
A SchedulerProcess instance containing the center-aligned scheduling strategy.
"""




#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
There are generally two ways in experiments for characterizing the performance of a quantum
computer in a superconducting platform: Quantum Process Tomography(QPT) and Randomized Benchmarking (RB)
. QPT can completely characterize a gate, decomposing a process into Pauli or Kraus operators,
but improving gates by QPT is complicated and resource-consuming. Also, State Preparation And Measurement
(SPAM) errors can be confused with process errors. However, RB is a concept of using randomization
methods for benchmarking quantum gates. It is a scalable and SPAM robust method for benchmarking the
full set of gates by a single parameter using randomization techniques. So it is useful to use RB for
a relatively simple and at least SPAM-robust benchmark, especially when the number of qubit increases.

For more details, please visit https://quanlse.baidu.com/#/doc/tutorial-randomized-benchmarking.
"""

import numpy as np
from numpy import ndarray
from functools import reduce

from Quanlse.QOperation.RotationGate import U
from Quanlse.QWaveform import QJob
from Quanlse.remoteSimulator import remoteSimulatorRunHamiltonian as runHamiltonian

from Quanlse.Utils.Functions import fromMatrixToAngles
from Quanlse.Utils.Clifford import clifford1q
from Quanlse.QPlatform import Error
from Quanlse.Simulator import PulseModel
from Quanlse.Scheduler.Superconduct import SchedulerSuperconduct
from Quanlse.QOperation.FixedGate import FixedGateOP
from Quanlse.Utils.Infidelity import rhoInfidelity


def RB(model: PulseModel, targetQubitNum: int, initialState: ndarray, size: int, width: int,
       sche: SchedulerSuperconduct, dt: float, targetGate: FixedGateOP = None, interleaved: bool = False,
       isOpen: bool = False) -> float:
    r"""
    Return the sequence's average fidelity for different number of Clifford.

    :param model: the QHamiltonian object of the multi-qubit system.
    :param targetQubitNum: the index of the qubit being benchmarked.
    :param initialState: the initial state of the system.
    :param size: the number of Cliffords in a "width".
    :param width: the number of random sequence(s) for each "size".
    :param sche: the initiated scheduler.
    :param dt: AWG's sampling time.
    :param targetGate: the gate being benchmarked.
    :param interleaved: whether to use the interleaved benchmarking method.
    :param isOpen: whether to solve open system.
    :return: the average sequence fidelity of different number of Cliffords.
    """
    # Test inputs correctness
    if not isinstance(targetQubitNum, int):
        raise Error.ArgumentError("We only support single-qubit RB for now.")
    else:
        job = QJob(subSysNum=model.subSysNum, sysLevel=model.sysLevel, dt=dt)
        ham = model.createQHamiltonian()
        jobs = ham.createJobList()
        # Get the RB-target qubit
        ham.job = job
        hamTarget = ham.subSystem(targetQubitNum)
        sche.clearCircuit()
        finalFidelity = []
        # Generate random Clifford
        a = 0
        while a < width:
            idx = np.random.randint(0, 24, size)
            b = 0
            clifford1qInvReg = []
            # Schedule m Clifford gates
            while b < size:
                if not interleaved:
                    clifford1q[idx[size - b - 1]](sche.Q[0])
                    clifford1qInvReg.append(clifford1q[idx[b]].getMatrix())
                else:
                    targetGate(sche.Q[0])
                    clifford1q[idx[size - b - 1]](sche.Q[0])
                    clifford1qInvReg.append(clifford1q[idx[b]].getMatrix() @ targetGate.getMatrix())
                b += 1
            # Schedule inverse Clifford
            gateUnitary = np.transpose(np.conjugate(reduce(np.dot, clifford1qInvReg)))
            [_, theta, phi, lamda] = fromMatrixToAngles(gateUnitary)
            U(theta, phi, lamda)(sche.Q[0])
            # Schedule the whole RB sequence
            jobTarget = sche.schedule()
            hamTarget.job = jobTarget
            jobScheFinal = hamTarget.outputInverseJob(subSysNum=model.subSysNum, sysLevel=model.sysLevel, dt=dt)
            job = model.getSimJob(jobScheFinal)
            jobs.addJob(jobs=job)
            sche.clearCircuit()
            jobTarget.clearWaves()
            jobScheFinal.clearWaves()
            job.clearWaves()
            a += 1
        ham.buildCache()
        # Solve the ODE function
        if not isOpen:
            result = runHamiltonian(ham, state0=initialState, jobList=jobs)
        else:
            result = runHamiltonian(ham, state0=initialState, jobList=jobs, isOpen=True)
        rhoi = initialState @ ((initialState.transpose()).conjugate())
        for res in result.result:
            psit = res['state']
            rhot = psit @ ((psit.transpose()).conjugate())
            # Calculate the single sequence fidelity
            fidelity = 1 - np.real(rhoInfidelity(rhot, rhoi))
            finalFidelity.append(fidelity)
        ham.job.clearWaves()
        ham.clearCache()
        hamTarget.job.clearWaves()
        hamTarget.clearCache()
    return sum(finalFidelity) / width
